from __future__ import annotations

import logging
import sys
from datetime import datetime
from typing import Any

from airflow.configuration import secrets_backend_list
from airflow.exceptions import AirflowSkipException
from airflow.models.dag import DAG
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.secrets.local_filesystem import LocalFilesystemBackend
from airflow.utils import timezone
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.utils.state import DagRunState, State
from airflow.utils.types import DagRunType
from sqlalchemy.orm.session import Session

log = logging.getLogger(__name__)


def run_dag(dag: DAG, conn_file_path: str) -> DagRun:
    return test_dag(dag=dag, conn_file_path=conn_file_path)


# DAG.test() was added in Airflow version 2.5.0. And to test on older Airflow versions, we need to copy the
# implementation here.
@provide_session
def test_dag(
    dag,
    execution_date: datetime | None = None,
    run_conf: dict[str, Any] | None = None,
    conn_file_path: str | None = None,
    variable_file_path: str | None = None,
    session: Session = NEW_SESSION,
) -> DagRun:
    """
    Execute one single DagRun for a given DAG and execution date.

    :param execution_date: execution date for the DAG run
    :param run_conf: configuration to pass to newly created dagrun
    :param conn_file_path: file path to a connection file in either yaml or json
    :param variable_file_path: file path to a variable file in either yaml or json
    :param session: database connection (optional)
    """

    if conn_file_path or variable_file_path:
        local_secrets = LocalFilesystemBackend(
            variables_file_path=variable_file_path, connections_file_path=conn_file_path
        )
        secrets_backend_list.insert(0, local_secrets)

    execution_date = execution_date or timezone.utcnow()

    dag.log.debug(
        "Clearing existing task instances for execution date %s", execution_date
    )
    dag.clear(
        start_date=execution_date,
        end_date=execution_date,
        dag_run_state=False,  # type: ignore
        session=session,
    )
    dag.log.debug("Getting dagrun for dag %s", dag.dag_id)
    dr: DagRun = _get_or_create_dagrun(
        dag=dag,
        start_date=execution_date,
        execution_date=execution_date,
        run_id=DagRun.generate_run_id(DagRunType.MANUAL, execution_date),
        session=session,
        conf=run_conf,
    )

    tasks = dag.task_dict
    dag.log.debug("starting dagrun")
    # Instead of starting a scheduler, we run the minimal loop possible to check
    # for task readiness and dependency management. This is notably faster
    # than creating a BackfillJob and allows us to surface logs to the user
    while dr.state == State.RUNNING:
        schedulable_tis, _ = dr.update_state(session=session)
        for ti in schedulable_tis:
            add_logger_if_needed(dag, ti)
            ti.task = tasks[ti.task_id]
            _run_task(ti, session=session)
    if conn_file_path or variable_file_path:
        # Remove the local variables we have added to the secrets_backend_list
        secrets_backend_list.pop(0)

    print("conn_file_path", conn_file_path)

    return dr


def add_logger_if_needed(dag: DAG, ti: TaskInstance):
    """
    Add a formatted logger to the taskinstance so all logs are surfaced to the command line instead
    of into a task file. Since this is a local test run, it is much better for the user to see logs
    in the command line, rather than needing to search for a log file.
    Args:
        ti: The taskinstance that will receive a logger

    """
    logging_format = logging.Formatter(
        "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    )
    handler = logging.StreamHandler(sys.stdout)
    handler.level = logging.INFO
    handler.setFormatter(logging_format)
    # only add log handler once
    if not any(isinstance(h, logging.StreamHandler) for h in ti.log.handlers):
        dag.log.debug("Adding Streamhandler to taskinstance %s", ti.task_id)
        ti.log.addHandler(handler)


def _run_task(ti: TaskInstance, session):
    """
    Run a single task instance, and push result to Xcom for downstream tasks. Bypasses a lot of
    extra steps used in `task.run` to keep our local running as fast as possible
    This function is only meant for the `dag.test` function as a helper function.

    Args:
        ti: TaskInstance to run
    """
    log.info("*****************************************************")
    if hasattr(ti, "map_index") and ti.map_index > 0:
        log.info("Running task %s index %d", ti.task_id, ti.map_index)
    else:
        log.info("Running task %s", ti.task_id)
    try:
        ti._run_raw_task(session=session)
        session.flush()
        log.info("%s ran successfully!", ti.task_id)
    except AirflowSkipException:
        log.info("Task Skipped, continuing")
    log.info("*****************************************************")


def _get_or_create_dagrun(
    dag: DAG,
    conf: dict[Any, Any] | None,
    start_date: datetime,
    execution_date: datetime,
    run_id: str,
    session: Session,
) -> DagRun:
    """
    Create a DAGRun, but only after clearing the previous instance of said dagrun to prevent collisions.
    This function is only meant for the `dag.test` function as a helper function.
    :param dag: Dag to be used to find dagrun
    :param conf: configuration to pass to newly created dagrun
    :param start_date: start date of new dagrun, defaults to execution_date
    :param execution_date: execution_date for finding the dagrun
    :param run_id: run_id to pass to new dagrun
    :param session: sqlalchemy session
    :return:
    """
    log.info("dagrun id: %s", dag.dag_id)
    dr: DagRun = (
        session.query(DagRun)
        .filter(DagRun.dag_id == dag.dag_id, DagRun.execution_date == execution_date)
        .first()
    )
    if dr:
        session.delete(dr)
        session.commit()
    dr = dag.create_dagrun(
        state=DagRunState.RUNNING,
        execution_date=execution_date,
        run_id=run_id,
        start_date=start_date or execution_date,
        session=session,
        conf=conf,  # type: ignore
    )
    log.info("created dagrun %s", str(dr))
    return dr



import os
import sys

# Add the project root to the path so we can import the package
sys.path.insert(0, os.path.abspath("../"))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Astronomer Cosmos"
copyright = "2023, Astronomer"
author = "Astronomer"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx_tabs.tabs",
]

autosummary_generate = True
add_module_names = False
autodoc_mock_imports = [
    "airflow",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]
html_theme_options = {
    "logo": {
        "image_light": "logo-light.png",
        "image_dark": "logo-dark.png",
    },
    "footer_items": ["copyright"],
}



"""Example DAG for using the DatabricksWorkflowTaskGroup and DatabricksNotebookOperator."""
import os
from datetime import timedelta

from airflow.models.dag import DAG
from airflow.utils.timezone import datetime

from astro_databricks import DatabricksNotebookOperator, DatabricksWorkflowTaskGroup

EXECUTION_TIMEOUT = int(os.getenv("EXECUTION_TIMEOUT", 6))
default_args = {
    "execution_timeout": timedelta(hours=EXECUTION_TIMEOUT),
    "retries": int(os.getenv("DEFAULT_TASK_RETRIES", 2)),
    "retry_delay": timedelta(seconds=int(os.getenv("DEFAULT_RETRY_DELAY_SECONDS", 60))),
}

DATABRICKS_CONN_ID = os.getenv("ASTRO_DATABRICKS_CONN_ID", "databricks_conn")
DATABRICKS_NOTIFICATION_EMAIL = os.getenv(
    "ASTRO_DATABRICKS_NOTIFICATION_EMAIL", "tatiana.alchueyr@astronomer.io"
)
DATABRICKS_DESTINATION_ID = os.getenv(
    "ASTRO_DATABRICKS_DESTINATION_ID", "b0aea8ab-ea8c-4a45-a2e9-9a26753fd702"
)
GROUP_ID = os.getenv("DATABRICKS_GROUP_ID", "1234").replace(".", "_")
job_cluster_spec = [
    {
        "job_cluster_key": "Shared_job_cluster",
        "new_cluster": {
            "cluster_name": "",
            "spark_version": "11.3.x-scala2.12",
            "aws_attributes": {
                "first_on_demand": 1,
                "availability": "SPOT_WITH_FALLBACK",
                "zone_id": "us-east-2b",
                "spot_bid_price_percent": 100,
                "ebs_volume_count": 0,
            },
            "node_type_id": "i3.xlarge",
            "spark_env_vars": {"PYSPARK_PYTHON": "/databricks/python3/bin/python3"},
            "enable_elastic_disk": False,
            "data_security_mode": "LEGACY_SINGLE_USER_STANDARD",
            "runtime_engine": "STANDARD",
            "num_workers": 8,
        },
    }
]
dag = DAG(
    dag_id="example_databricks_workflow",
    start_date=datetime(2022, 1, 1),
    schedule_interval=None,
    catchup=False,
    default_args=default_args,
    tags=["example", "async", "databricks"],
)
with dag:
    # [START howto_databricks_workflow_notebook]
    task_group = DatabricksWorkflowTaskGroup(
        group_id=f"test_workflow_{GROUP_ID}",
        databricks_conn_id=DATABRICKS_CONN_ID,
        job_clusters=job_cluster_spec,
        notebook_params=[],
        notebook_packages=[
            {
                "pypi": {
                    "package": "simplejson==3.18.0",  # Pin specification version of a package like this.
                    "repo": "https://pypi.org/simple",  # You can specify your required Pypi index here.
                }
            },
        ],
        extra_job_params={
            "email_notifications": {
                "on_start": [DATABRICKS_NOTIFICATION_EMAIL],
            },
            "webhook_notifications": {
                "on_start": [{"id": DATABRICKS_DESTINATION_ID}],
            },
        },
    )
    with task_group:
        notebook_1 = DatabricksNotebookOperator(
            task_id="notebook_1",
            databricks_conn_id=DATABRICKS_CONN_ID,
            notebook_path="/Shared/Notebook_1",
            notebook_packages=[{"pypi": {"package": "Faker"}}],
            source="WORKSPACE",
            job_cluster_key="Shared_job_cluster",
            execution_timeout=timedelta(seconds=600),
        )
        notebook_2 = DatabricksNotebookOperator(
            task_id="notebook_2",
            databricks_conn_id=DATABRICKS_CONN_ID,
            notebook_path="/Shared/Notebook_2",
            source="WORKSPACE",
            job_cluster_key="Shared_job_cluster",
            notebook_params={
                "foo": "bar",
            },
        )
        notebook_1 >> notebook_2
    # [END howto_databricks_workflow_notebook]



from __future__ import annotations

import uuid
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest
from airflow import DAG
from airflow.models.dagbag import DagBag
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstanceKey
from airflow.providers.databricks.hooks.databricks import DatabricksHook
from airflow.utils.dates import days_ago
from airflow.utils.db import create_session
from airflow.utils.state import State
from astro_databricks.plugins.plugin import (
    DatabricksJobRepairAllFailedLink,
    DatabricksJobRepairSingleFailedLink,
    DatabricksJobRunLink,
    _clear_task_instances,
    _get_dagrun,
    _get_databricks_task_id,
    _repair_task,
)
from databricks_cli.sdk.service import JobsService


@pytest.fixture
def mock_dag():
    dag = MagicMock(spec=DAG)
    dag.dag_id = "my_dag"
    dag.get_task.return_value = MagicMock(
        task_group=MagicMock(
            children={"task_1": "task1", "task_2": "task2", "task_5": "task5"},
            group_id="test_group",
        )
    )
    return dag


@pytest.fixture
def mock_session():
    return MagicMock()


@patch("astro_databricks.plugins.plugin.get_airflow_app")
@patch("astro_databricks.plugins.plugin._get_dagrun")
@patch("astro_databricks.plugins.plugin.clear_task_instances")
def test_clear_task_instances(
    mock_clear_tasks, mock_dagrun, mock_get_airflow_app, session
):
    mock_get_airflow_app.return_value.dag_bag = MagicMock()

    dag_id = "test_dag"
    run_id = "test_run"
    task_ids = ["task_1", "task_2", "task_3"]

    # Create mock objects for testing
    dag = MagicMock()
    dag.get_task_instances.return_value = [
        MagicMock(task_id=task_id, dag_id=dag_id) for task_id in task_ids
    ]

    dag_run = MagicMock(spec=DagRun)
    dag_run.get_task_instances.return_value = dag.get_task_instances()
    mock_dagrun.return_value = dag_run
    mock_session = MagicMock()
    databricks_task_ids = [dag_id + "__" + task_id for task_id in task_ids]
    _clear_task_instances(
        dag_id, run_id, databricks_task_ids, log=MagicMock(), session=mock_session
    )

    # Assert that clear_task_instances was called with the correct list of TaskInstances
    expected_tis = dag_run.get_task_instances.return_value
    mock_clear_tasks.assert_called_once_with(expected_tis, mock_session)


@patch("astro_databricks.plugins.plugin.get_airflow_app")
@patch("astro_databricks.plugins.plugin.XCom.get_value")
@patch("astro_databricks.plugins.plugin.DatabricksHook")
@patch("astro_databricks.plugins.plugin.get_task_group")
def test_databricks_job_run_link(
    mock_get_task_group, mock_hook, mock_xcom, mock_get_airflow_app, mock_dag
):
    mock_dag_bag = MagicMock()
    mock_dag_bag.get_dag.return_value = mock_dag
    mock_get_airflow_app.return_value.dag_bag = mock_dag_bag

    mock_xcom.return_value = {
        "databricks_job_id": "test_job",
        "databricks_run_id": "test_run",
        "databricks_conn_id": "test_conn",
    }

    mock_hook.return_value.host = "test_host"
    mock_get_task_group.return_value.get_child_by_label.return_value.task_id = (
        "test_group.launch"
    )

    link = DatabricksJobRunLink()
    operator = MagicMock(
        task_id="dummy_task",
        dag=mock_dag,
        task_group=MagicMock(group_id="test_group", default_args={}),
    )
    ti_key = TaskInstanceKey(dag_id="test_dag", task_id="dummy_task", run_id="test_run")
    result = link.get_link(operator=operator, ti_key=ti_key)

    mock_dag_bag.get_dag.assert_called_once_with("test_dag")
    mock_xcom.assert_called_once_with(
        ti_key=TaskInstanceKey(
            task_id="test_group.launch", dag_id="test_dag", run_id="test_run"
        ),
        key="return_value",
    )
    mock_hook.assert_called_once_with("test_conn")

    expected_result = "https://test_host/#job/test_job/run/test_run"
    assert result == expected_result


@pytest.fixture
def mock_dagrun():
    dagrun = MagicMock(spec=DagRun)
    dagrun.get_task_instances.return_value = [
        MagicMock(state="failed", task_id="task_1"),
        MagicMock(state="skipped", task_id="task_2"),
        MagicMock(state="up_for_retry", task_id="task_3"),
        MagicMock(state="upstream_failed", task_id="task_4"),
        MagicMock(state="success", task_id="task_5"),
    ]
    return dagrun


@mock.patch("astro_databricks.plugins.plugin.XCom")
def test_repair_all_get_link(mock_xcom, mock_dagrun, mock_dag, mock_session):
    # Arrange
    task_instance_key = TaskInstanceKey(
        dag_id="my_dag",
        task_id="my_task",
        run_id="run_id",
    )
    mock_xcom.get_value.return_value = {
        "databricks_job_id": "job_id",
        "databricks_run_id": "run_id",
        "databricks_conn_id": "databricks_conn",
    }

    link = DatabricksJobRepairAllFailedLink()

    link.get_dagrun = MagicMock(return_value=mock_dagrun)
    link.get_dag = MagicMock(return_value=mock_dag)
    link.get_tasks_to_run = MagicMock(return_value="task_1,task_2")
    mock_operator = MagicMock(task_group=MagicMock(group_id="test_group"))
    # Act
    result = link.get_link(mock_operator, None, ti_key=task_instance_key)

    # Assert
    assert (
        result == "/repair_databricks_job?dag_id=my_dag&"
        "databricks_conn_id=databricks_conn&"
        "databricks_run_id=run_id&"
        "run_id=run_id&"
        "tasks_to_repair=task_1,task_2"
    )


@mock.patch("astro_databricks.plugins.plugin.get_airflow_app")
@mock.patch("astro_databricks.plugins.plugin._get_dagrun")
def test_get_tasks_to_run(mock_dagrun, mock_airflow_app):
    link = DatabricksJobRepairAllFailedLink()
    ti_key = TaskInstanceKey(dag_id="test_dag", task_id="test_task", run_id="test_run")
    dag = DAG("test_dag")
    task_group_children = {
        "test_group.test_task": MagicMock(
            task_id="test_group.test_task", dag_id="test_dag"
        ),
        "test_group.test_task_2": MagicMock(
            task_id="test_group.test_task_2", dag_id="test_dag"
        ),
        "test_group.test_task_3": MagicMock(
            task_id="test_group.test_task_3", dag_id="test_dag"
        ),
    }
    task = MagicMock(
        task_id="test_task",
        task_group=MagicMock(
            dag_id="test_dag", group_id="test_group", children=task_group_children
        ),
    )
    mock_airflow_app.return_value.dag_bag.get_dag.return_value = dag
    dag.add_task(task)

    def generate_mock_dagrun(task_map: dict[str, str]):
        dagrun = MagicMock(spec=DagRun)
        dagrun.get_task_instances.return_value = [
            MagicMock(state=state, task_id=task_id)
            for task_id, state in task_map.items()
        ]
        return dagrun

    # Case 1: No failed or skipped tasks
    mock_dagrun.return_value = generate_mock_dagrun({})
    tasks_str = link.get_tasks_to_run(ti_key, operator=task, log=MagicMock())
    assert tasks_str == ""

    # # Case 2: One failed task
    mock_dagrun.return_value = generate_mock_dagrun({"test_group.test_task": "failed"})
    tasks_str = link.get_tasks_to_run(ti_key, task, log=MagicMock())
    assert tasks_str == "test_dag__test_group__test_task"
    #
    # # Case 3: One skipped task
    mock_dagrun.return_value = generate_mock_dagrun({"test_group.test_task": "skipped"})
    tasks_str = link.get_tasks_to_run(ti_key, task, log=MagicMock())
    assert tasks_str == "test_dag__test_group__test_task"
    #
    # # Case 4: Multiple failed and skipped tasks
    mock_dagrun.return_value = generate_mock_dagrun(
        {"test_group.test_task": "failed", "test_group.test_task_2": "skipped"}
    )
    tasks_str = link.get_tasks_to_run(ti_key, task, log=MagicMock())
    assert (
        tasks_str == "test_dag__test_group__test_task,test_dag__test_group__test_task_2"
    )


@pytest.fixture
def session():
    with create_session() as session:
        yield session


def test_get_dagrun(session, dag):
    # Create a DagRun object and add it to the database
    DagBag()
    run_id = "test_run_id" + uuid.uuid4().hex
    dr = dag.create_dagrun(run_id=run_id, state=State.RUNNING)
    session.add(dr)
    session.commit()

    # Call the function and ensure it returns the correct DagRun object
    result = _get_dagrun(dag, run_id, session=session)
    assert result == dr


@mock.patch("astro_databricks.plugins.plugin.DatabricksHook")
@mock.patch("astro_databricks.plugins.plugin.JobsService")
@mock.patch("astro_databricks.plugins.plugin.ApiClient")
def test_repair_task(mock_api_client, mock_jobs_service, mock_hook):
    databricks_conn_id = "my_databricks_conn"
    databricks_run_id = "my_databricks_run"
    tasks_to_repair = ["task1", "task2"]

    # Mock the Databricks hook and API client
    mock_hook.return_value = MagicMock(spec=DatabricksHook)
    mock_api_client = MagicMock()
    mock_hook.get_conn.return_value = mock_api_client

    # Mock the JobsService and its methods
    mock_jobs_service.return_value = MagicMock(spec=JobsService)
    mock_jobs_service.return_value.get_run.return_value = {
        "job_id": 1234,
        "run_id": databricks_run_id,
        "state": "RUNNING",
        "start_time": "2022-02-27T00:00:00Z",
        "end_time": None,
        "tasks": [],
        "state_message": None,
        "creator_user_name": "airflow",
        "run_name": None,
        "run_page_url": None,
        "run_type": None,
        "spark_context_id": None,
        "retry_number": 0,
        "previous_run_id": None,
        "trigger": {},
        "is_completed": False,
        "is_active": True,
        "is_queued": False,
        "cluster_spec": {},
        "overriding_parameters": {},
        "start_time_epoch": 1645958400,
    }
    mock_jobs_service.return_value.repair.return_value = None

    # Patch the DatabricksHook and JobsService constructors
    _repair_task(
        databricks_conn_id, databricks_run_id, tasks_to_repair, log=MagicMock()
    )

    # # Check that the JobsService methods were called correctly
    mock_hook.return_value.get_conn.assert_called_once_with()
    mock_jobs_service.return_value.get_run.assert_called_once_with(
        run_id=databricks_run_id, include_history=True
    )
    mock_jobs_service.return_value.repair.assert_called_once_with(
        run_id=databricks_run_id,
        version="2.1",
        latest_repair_id=None,
        rerun_tasks=tasks_to_repair,
    )


@patch("astro_databricks.plugins.plugin.get_airflow_app")
def test_databricks_job_repair_single_failed_link(mock_get_airflow_app, dag):
    mock_dag_bag = MagicMock()
    mock_task = MagicMock(
        task_id="test_task",
        task_group=MagicMock(group_id="test_group", default_args={}),
    )
    test_dag = DAG("test_dag", start_date=days_ago(1))
    mock_dag_bag.get_dag.return_value = test_dag
    test_dag.get_task = MagicMock(return_value=mock_task)
    mock_get_airflow_app.return_value.dag_bag = mock_dag_bag
    link = DatabricksJobRepairSingleFailedLink()

    dag_id = "test_dag"
    task_id = "test_task"
    run_id = "test_run"
    databricks_conn_id = "test_conn"
    databricks_run_id = "test_run_id"

    ti_key = TaskInstanceKey(dag_id, task_id, run_id)
    metadata = {
        "databricks_conn_id": databricks_conn_id,
        "databricks_run_id": databricks_run_id,
        "databricks_job_id": 1234,
    }

    mock_xcom = MagicMock()
    mock_xcom.get_value.return_value = metadata

    with patch("astro_databricks.plugins.plugin.XCom", mock_xcom):
        link.get_link(mock_task, dttm=None, ti_key=ti_key)
        f"/repair_databricks_job?dag_id={dag_id}&databricks_conn_id={databricks_conn_id}&databricks_run_id={databricks_run_id}&tasks_to_repair={_get_databricks_task_id(mock_task)}"



def get_job_run_url(host: str, job_id: str, run_id: str) -> str:
    """
    Return Databricks Job Run URL.
    :param host: Databricks host
    :param job_id: Databricks Workflow Job Identifier
    :param run_id: Databricks Workflow Job Run Identifier

    :return: URL which the user can click to visualise the job in the Databricks UI.
    """
    return f"https://{host}/#job/{job_id}/run/{run_id}"


import pytest
from airflow.models import DAG

try:
    from airflow.operators.empty import EmptyOperator
except ModuleNotFoundError:
    # For Airflow versions < 2.3 which do not contain the EmptyOperator.
    from airflow.operators.dummy import DummyOperator as EmptyOperator

from airflow.utils.timezone import datetime

TEST_DAG_ID = "unit_test_dag"


def pytest_addoption(parser):
    parser.addoption("--package", action="store")


@pytest.fixture(scope="session")
def package(request):
    package_value = request.config.option.package
    if package_value is None:
        pytest.skip()
    return package_value


@pytest.fixture
def dag():
    """
    Creates a test DAG with default arguments.
    """
    dag = DAG(TEST_DAG_ID, start_date=datetime(2022, 1, 1))
    with dag:
        EmptyOperator(task_id="foo")
    yield dag



#!/usr/bin/env python3
"""Verify the version of the Package with the version in Git tag."""

import os
import re
from pathlib import Path

repo_dir = Path(__file__).parent.parent.parent

# TODO: Update this to match your package name
path_of_init_file = Path(repo_dir / "great_expectations_provider" / "__init__.py")
version_file = path_of_init_file.read_text()
git_ref = os.getenv("GITHUB_REF", "")
git_tag = git_ref.replace("refs/tags/", "")
version = re.findall('__version__ = "(.*)"', version_file)[0]

if git_tag is not None:
    if version != git_tag:
        raise SystemExit(f"The version in {path_of_init_file} ({version}) does not match the Git Tag ({git_tag}).")



import json
from datetime import datetime, timedelta

from airflow.decorators import (
    dag,
    task,
)  # DAG and task decorators for interfacing with the TaskFlow API


@dag(
    # This defines how often your DAG will run, or the schedule by which your DAG runs. In this case, this DAG
    # will run daily
    schedule_interval="@daily",
    # This DAG is set to run for the first time on January 1, 2021. Best practice is to use a static
    # start_date. Subsequent DAG runs are instantiated based on scheduler_interval
    start_date=datetime(2021, 1, 1),
    # When catchup=False, your DAG will only run for the latest schedule_interval. In this case, this means
    # that tasks will not be run between January 1, 2021 and 30 mins ago. When turned on, this DAG's first
    # run will be for the next 30 mins, per the schedule_interval
    catchup=False,
    default_args={
        "retries": 2,  # If a task fails, it will retry 2 times.
    },
    tags=["example"],
)  # If set, this tag is shown in the DAG view of the Airflow UI
def example_dag_basic():
    """
    ### Basic ETL Dag
    This is a simple ETL data pipeline example that demonstrates the use of
    the TaskFlow API using three simple tasks for extract, transform, and load.
    For more information on Airflow's TaskFlow API, reference documentation here:
    https://airflow.apache.org/docs/apache-airflow/stable/tutorial_taskflow_api.html
    """

    @task()
    def extract():
        """
        #### Extract task
        A simple "extract" task to get data ready for the rest of the
        pipeline. In this case, getting data is simulated by reading from a
        hardcoded JSON string.
        """
        data_string = '{"1001": 301.27, "1002": 433.21, "1003": 502.22}'

        order_data_dict = json.loads(data_string)
        return order_data_dict

    @task(
        multiple_outputs=True
    )  # multiple_outputs=True unrolls dictionaries into separate XCom values
    def transform(order_data_dict: dict):
        """
        #### Transform task
        A simple "transform" task which takes in the collection of order data and
        computes the total order value.
        """
        total_order_value = 0

        for value in order_data_dict.values():
            total_order_value += value

        return {"total_order_value": total_order_value}

    @task()
    def load(total_order_value: float):
        """
        #### Load task
        A simple "load" task that takes in the result of the "transform" task and prints it out,
        instead of saving it to end user review
        """

        print(f"Total order value is: {total_order_value:.2f}")

    order_data = extract()
    order_summary = transform(order_data)
    load(order_summary["total_order_value"])


example_dag_basic = example_dag_basic()



from unittest import mock

from astro_databricks.operators.notebook import DatabricksNotebookOperator
from astro_databricks.operators.workflow import DatabricksWorkflowTaskGroup

expected_workflow_json = {
    "name": "unit_test_dag.test_workflow",
    "email_notifications": {"no_alert_for_skipped_runs": False},
    "format": "MULTI_TASK",
    "job_clusters": [{"job_cluster_key": "foo"}],
    "max_concurrent_runs": 1,
    "tasks": [
        {
            "depends_on": [],
            "email_notifications": {},
            "job_cluster_key": "foo",
            "libraries": [
                {"nb_index": {"package": "nb_package"}},
                {"tg_index": {"package": "tg_package"}},
            ],
            "notebook_task": {
                "base_parameters": {},
                "notebook_path": "/foo/bar",
                "source": "WORKSPACE",
            },
            "task_key": "unit_test_dag__test_workflow__notebook_1",
            "timeout_seconds": 0,
        },
        {
            "depends_on": [{"task_key": "unit_test_dag__test_workflow__notebook_1"}],
            "email_notifications": {},
            "job_cluster_key": "foo",
            "libraries": [{"tg_index": {"package": "tg_package"}}],
            "notebook_task": {
                "base_parameters": {"foo": "bar"},
                "notebook_path": "/foo/bar",
                "source": "WORKSPACE",
            },
            "task_key": "unit_test_dag__test_workflow__notebook_2",
            "timeout_seconds": 0,
        },
    ],
    "timeout_seconds": 0,
}


@mock.patch("astro_databricks.operators.workflow.DatabricksHook")
@mock.patch("astro_databricks.operators.workflow.ApiClient")
@mock.patch("astro_databricks.operators.workflow.JobsApi")
def test_create_workflow_from_notebooks_with_create(
    mock_jobs_api, mock_api, mock_hook, dag
):
    mock_jobs_api.return_value.create_job.return_value = {"job_id": 1}
    with dag:
        task_group = DatabricksWorkflowTaskGroup(
            group_id="test_workflow",
            databricks_conn_id="foo",
            job_clusters=[{"job_cluster_key": "foo"}],
            notebook_params=[{"notebook_path": "/foo/bar"}],
            notebook_packages=[{"tg_index": {"package": "tg_package"}}],
        )
        with task_group:
            notebook_1 = DatabricksNotebookOperator(
                task_id="notebook_1",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                notebook_packages=[{"nb_index": {"package": "nb_package"}}],
                source="WORKSPACE",
                job_cluster_key="foo",
            )
            notebook_2 = DatabricksNotebookOperator(
                task_id="notebook_2",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
                notebook_params={
                    "foo": "bar",
                },
            )
            notebook_1 >> notebook_2

    assert len(task_group.children) == 3
    task_group.children["test_workflow.launch"].execute(context={})
    mock_jobs_api.return_value.create_job.assert_called_once_with(
        json=expected_workflow_json,
    )
    mock_jobs_api.return_value.run_now.assert_called_once_with(
        job_id=1,
        jar_params=[],
        notebook_params=[{"notebook_path": "/foo/bar"}],
        python_params=[],
        spark_submit_params=[],
    )


@mock.patch("astro_databricks.operators.workflow.DatabricksHook")
@mock.patch("astro_databricks.operators.workflow.ApiClient")
@mock.patch("astro_databricks.operators.workflow.JobsApi")
@mock.patch("astro_databricks.operators.workflow._get_job_by_name")
def test_create_workflow_from_notebooks_existing_job(
    mock_get_jobs, mock_jobs_api, mock_api, mock_hook, dag
):
    mock_get_jobs.return_value = {"job_id": 1}
    with dag:
        task_group = DatabricksWorkflowTaskGroup(
            group_id="test_workflow",
            databricks_conn_id="foo",
            job_clusters=[{"job_cluster_key": "foo"}],
            notebook_params=[{"notebook_path": "/foo/bar"}],
            notebook_packages=[{"tg_index": {"package": "tg_package"}}],
        )
        with task_group:
            notebook_1 = DatabricksNotebookOperator(
                task_id="notebook_1",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                notebook_packages=[{"nb_index": {"package": "nb_package"}}],
                source="WORKSPACE",
                job_cluster_key="foo",
            )
            notebook_2 = DatabricksNotebookOperator(
                task_id="notebook_2",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
                notebook_params={
                    "foo": "bar",
                },
            )
            notebook_1 >> notebook_2

    assert len(task_group.children) == 3
    task_group.children["test_workflow.launch"].execute(context={})
    mock_jobs_api.return_value.reset_job.assert_called_once_with(
        json={"job_id": 1, "new_settings": expected_workflow_json},
    )
    mock_jobs_api.return_value.run_now.assert_called_once_with(
        job_id=1,
        jar_params=[],
        notebook_params=[{"notebook_path": "/foo/bar"}],
        python_params=[],
        spark_submit_params=[],
    )


@mock.patch("astro_databricks.operators.workflow.DatabricksHook")
@mock.patch("astro_databricks.operators.workflow.ApiClient")
@mock.patch("astro_databricks.operators.workflow.JobsApi")
@mock.patch("astro_databricks.operators.workflow._get_job_by_name")
def test_create_workflow_with_arbitrary_extra_job_params(
    mock_get_jobs, mock_jobs_api, mock_api, mock_hook, dag
):
    mock_get_jobs.return_value = {"job_id": 862519602273592}

    extra_job_params = {
        "timeout_seconds": 10,  # default: 0
        "webhook_notifications": {
            "on_failure": [{"id": "b0aea8ab-ea8c-4a45-a2e9-9a26753fd702"}],
        },
        "email_notifications": {
            "no_alert_for_skipped_runs": True,  # default: False
            "on_start": ["user.name@databricks.com"],
        },
        "git_source": {  # no default value
            "git_url": "https://github.com/astronomer/astro-provider-databricks",
            "git_provider": "gitHub",
            "git_branch": "main",
        },
    }
    with dag:
        task_group = DatabricksWorkflowTaskGroup(
            group_id="test_workflow",
            databricks_conn_id="foo",
            job_clusters=[{"job_cluster_key": "foo"}],
            notebook_params=[{"notebook_path": "/foo/bar"}],
            extra_job_params=extra_job_params,
        )
        with task_group:
            notebook_with_extra = DatabricksNotebookOperator(
                task_id="notebook_with_extra",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
            )
            notebook_with_extra

    assert len(task_group.children) == 2

    task_group.children["test_workflow.launch"].create_workflow_json()
    task_group.children["test_workflow.launch"].execute(context={})

    mock_jobs_api.return_value.reset_job.assert_called_once()
    kwargs = mock_jobs_api.return_value.reset_job.call_args_list[0].kwargs["json"]

    assert kwargs["job_id"] == 862519602273592
    assert (
        kwargs["new_settings"]["email_notifications"]
        == extra_job_params["email_notifications"]
    )
    assert (
        kwargs["new_settings"]["timeout_seconds"] == extra_job_params["timeout_seconds"]
    )
    assert kwargs["new_settings"]["git_source"] == extra_job_params["git_source"]
    assert (
        kwargs["new_settings"]["webhook_notifications"]
        == extra_job_params["webhook_notifications"]
    )
    assert (
        kwargs["new_settings"]["email_notifications"]
        == extra_job_params["email_notifications"]
    )



from unittest import mock
from unittest.mock import MagicMock

import pytest
from airflow.exceptions import AirflowException
from astro_databricks.operators.notebook import DatabricksNotebookOperator
from astro_databricks.operators.workflow import (
    DatabricksWorkflowTaskGroup,
)


@pytest.fixture
def mock_runs_api():
    return MagicMock()


@pytest.fixture
def databricks_notebook_operator():
    return DatabricksNotebookOperator(
        task_id="notebook",
        databricks_conn_id="foo",
        notebook_path="/foo/bar",
        source="WORKSPACE",
        job_cluster_key="foo",
        notebook_params={
            "foo": "bar",
        },
        notebook_packages=[{"nb_index": {"package": "nb_package"}}],
    )


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.launch_notebook_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
def test_databricks_notebook_operator_without_taskgroup(mock_monitor, mock_launch, dag):
    with dag:
        notebook = DatabricksNotebookOperator(
            task_id="notebook",
            databricks_conn_id="foo",
            notebook_path="/foo/bar",
            source="WORKSPACE",
            job_cluster_key="foo",
            notebook_params={
                "foo": "bar",
            },
            notebook_packages=[{"nb_index": {"package": "nb_package"}}],
        )
        assert notebook.task_id == "notebook"
        assert notebook.databricks_conn_id == "foo"
        assert notebook.notebook_path == "/foo/bar"
        assert notebook.source == "WORKSPACE"
        assert notebook.job_cluster_key == "foo"
        assert notebook.notebook_params == {"foo": "bar"}
        assert notebook.notebook_packages == [{"nb_index": {"package": "nb_package"}}]
    dag.test()
    mock_launch.assert_called_once()
    mock_monitor.assert_called_once()


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.launch_notebook_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
@mock.patch(
    "astro_databricks.operators.workflow._CreateDatabricksWorkflowOperator.execute"
)
def test_databricks_notebook_operator_with_taskgroup(
    mock_create, mock_monitor, mock_launch, dag
):
    mock_create.return_value = {
        "databricks_job_id": "job_id",
        "databricks_run_id": "run_id",
        "databricks_conn_id": "conn_id",
    }
    with dag:
        task_group = DatabricksWorkflowTaskGroup(
            group_id="test_workflow",
            databricks_conn_id="foo",
            job_clusters=[{"job_cluster_key": "foo"}],
            notebook_params=[{"notebook_path": "/foo/bar"}],
        )
        with task_group:
            notebook = DatabricksNotebookOperator(
                task_id="notebook",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
                notebook_params={
                    "foo": "bar",
                },
                notebook_packages=[{"nb_index": {"package": "nb_package"}}],
            )
            assert notebook.task_id == "test_workflow.notebook"
            assert notebook.databricks_conn_id == "foo"
            assert notebook.notebook_path == "/foo/bar"
            assert notebook.source == "WORKSPACE"
            assert notebook.job_cluster_key == "foo"
            assert notebook.notebook_params == {"foo": "bar"}
            assert notebook.notebook_packages == [
                {"nb_index": {"package": "nb_package"}}
            ]
    dag.test()
    mock_launch.assert_not_called()
    mock_monitor.assert_called_once()


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_databricks_task_id"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch("astro_databricks.operators.notebook.RunsApi")
def test_databricks_notebook_operator_without_taskgroup_new_cluster(
    mock_runs_api, mock_api_client, mock_get_databricks_task_id, mock_monitor, dag
):
    mock_get_databricks_task_id.return_value = "1234"
    mock_runs_api.return_value = mock.MagicMock()
    with dag:
        DatabricksNotebookOperator(
            task_id="notebook",
            databricks_conn_id="foo",
            notebook_path="/foo/bar",
            source="WORKSPACE",
            job_cluster_key="foo",
            notebook_params={
                "foo": "bar",
            },
            notebook_packages=[{"nb_index": {"package": "nb_package"}}],
            new_cluster={"foo": "bar"},
        )
    dag.test()
    mock_runs_api.return_value.submit_run.assert_called_once_with(
        {
            "run_name": "1234",
            "notebook_task": {
                "notebook_path": "/foo/bar",
                "source": "WORKSPACE",
                "base_parameters": {"foo": "bar"},
            },
            "new_cluster": {"foo": "bar"},
            "libraries": [{"nb_index": {"package": "nb_package"}}],
            "timeout_seconds": 0,
            "email_notifications": {},
        }
    )
    mock_monitor.assert_called_once()


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_databricks_task_id"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch("astro_databricks.operators.notebook.RunsApi")
def test_databricks_notebook_operator_without_taskgroup_existing_cluster(
    mock_runs_api, mock_api_client, mock_get_databricks_task_id, mock_monitor, dag
):
    mock_get_databricks_task_id.return_value = "1234"
    mock_runs_api.return_value = mock.MagicMock()
    with dag:
        DatabricksNotebookOperator(
            task_id="notebook",
            databricks_conn_id="foo",
            notebook_path="/foo/bar",
            source="WORKSPACE",
            job_cluster_key="foo",
            notebook_params={
                "foo": "bar",
            },
            notebook_packages=[{"nb_index": {"package": "nb_package"}}],
            existing_cluster_id="123",
        )
    dag.test()
    mock_runs_api.return_value.submit_run.assert_called_once_with(
        {
            "run_name": "1234",
            "notebook_task": {
                "notebook_path": "/foo/bar",
                "source": "WORKSPACE",
                "base_parameters": {"foo": "bar"},
            },
            "existing_cluster_id": "123",
            "libraries": [{"nb_index": {"package": "nb_package"}}],
            "timeout_seconds": 0,
            "email_notifications": {},
        }
    )
    mock_monitor.assert_called_once()


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch("astro_databricks.operators.notebook.RunsApi")
def test_databricks_notebook_operator_without_taskgroup_existing_cluster_and_new_cluster(
    mock_runs_api, mock_api_client, mock_monitor, dag
):
    with pytest.raises(ValueError):
        with dag:
            DatabricksNotebookOperator(
                task_id="notebook",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
                notebook_params={
                    "foo": "bar",
                },
                notebook_packages=[{"nb_index": {"package": "nb_package"}}],
                existing_cluster_id="123",
                new_cluster={"foo": "bar"},
            )
        dag.test()


@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator.monitor_databricks_job"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch("astro_databricks.operators.notebook.RunsApi")
def test_databricks_notebook_operator_without_taskgroup_no_cluster(
    mock_runs_api, mock_api_client, mock_monitor, dag
):
    with pytest.raises(ValueError):
        with dag:
            DatabricksNotebookOperator(
                task_id="notebook",
                databricks_conn_id="foo",
                notebook_path="/foo/bar",
                source="WORKSPACE",
                job_cluster_key="foo",
                notebook_params={
                    "foo": "bar",
                },
                notebook_packages=[{"nb_index": {"package": "nb_package"}}],
            )
        dag.test()


def test_handle_final_state_success(databricks_notebook_operator):
    final_state = {
        "life_cycle_state": "TERMINATED",
        "result_state": "SUCCESS",
        "state_message": "Job succeeded",
    }
    databricks_notebook_operator._handle_final_state(final_state)


def test_handle_final_state_failure(databricks_notebook_operator):
    final_state = {
        "life_cycle_state": "TERMINATED",
        "result_state": "FAILED",
        "state_message": "Job failed",
    }
    with pytest.raises(AirflowException):
        databricks_notebook_operator._handle_final_state(final_state)


def test_handle_final_state_exception(databricks_notebook_operator):
    final_state = {
        "life_cycle_state": "SKIPPED",
        "state_message": "Job skipped",
    }
    with pytest.raises(AirflowException):
        databricks_notebook_operator._handle_final_state(final_state)


@mock.patch("astro_databricks.operators.notebook.RunsApi")
@mock.patch("time.sleep")
def test_wait_for_pending_task(mock_sleep, mock_runs_api, databricks_notebook_operator):
    # create a mock current task with "PENDING" state
    current_task = {"run_id": "123", "state": {"life_cycle_state": "PENDING"}}
    mock_runs_api.get_run.side_effect = [
        {"state": {"life_cycle_state": "PENDING"}},
        {"state": {"life_cycle_state": "RUNNING"}},
    ]
    databricks_notebook_operator._wait_for_pending_task(current_task, mock_runs_api)
    mock_runs_api.get_run.assert_called_with("123", version="2.1")
    assert mock_runs_api.get_run.call_count == 2
    mock_runs_api.reset_mock()


@mock.patch("astro_databricks.operators.notebook.RunsApi")
@mock.patch("time.sleep")
def test_wait_for_terminating_task(
    mock_sleep, mock_runs_api, databricks_notebook_operator
):
    current_task = {"run_id": "123", "state": {"life_cycle_state": "PENDING"}}
    mock_runs_api.get_run.side_effect = [
        {"state": {"life_cycle_state": "TERMINATING"}},
        {"state": {"life_cycle_state": "TERMINATING"}},
        {"state": {"life_cycle_state": "TERMINATED"}},
    ]
    databricks_notebook_operator._wait_for_terminating_task(current_task, mock_runs_api)
    mock_runs_api.get_run.assert_called_with("123", version="2.1")
    assert mock_runs_api.get_run.call_count == 3
    mock_runs_api.reset_mock()


@mock.patch("astro_databricks.operators.notebook.RunsApi")
@mock.patch("time.sleep")
def test_wait_for_running_task(mock_sleep, mock_runs_api, databricks_notebook_operator):
    current_task = {"run_id": "123", "state": {"life_cycle_state": "PENDING"}}
    mock_runs_api.get_run.side_effect = [
        {"state": {"life_cycle_state": "RUNNING"}},
        {"state": {"life_cycle_state": "RUNNING"}},
        {"state": {"life_cycle_state": "TERMINATED"}},
    ]
    databricks_notebook_operator._wait_for_running_task(current_task, mock_runs_api)
    mock_runs_api.get_run.assert_called_with("123", version="2.1")
    assert mock_runs_api.get_run.call_count == 3
    mock_runs_api.reset_mock()


def test_get_lifestyle_state(databricks_notebook_operator):
    runs_api_mock = MagicMock()
    runs_api_mock.get_run.return_value = {"state": {"life_cycle_state": "TERMINATING"}}

    task_info = {"run_id": "test_run_id"}

    assert (
        databricks_notebook_operator._get_lifestyle_state(task_info, runs_api_mock)
        == "TERMINATING"
    )


@mock.patch("astro_databricks.operators.notebook.DatabricksHook")
@mock.patch("astro_databricks.operators.notebook.RunsApi")
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_databricks_task_id"
)
def test_monitor_databricks_job_success(
    mock_get_databricks_task_id,
    mock_get_api_client,
    mock_runs_api,
    mock_databricks_hook,
    databricks_notebook_operator,
    caplog
):
    mock_get_databricks_task_id.return_value = "1"
    # Define the expected response
    response = {
        "run_page_url": "https://databricks-instance-xyz.cloud.databricks.com/#job/1234/run/1",
        "state": {
            "life_cycle_state": "TERMINATED",
            "result_state": "SUCCESS",
            "state_message": "Ran successfully",
        },
        "tasks": [
            {
                "run_id": "1",
                "task_key": "1",
            }
        ],
    }
    mock_runs_api.return_value.get_run.return_value = response

    databricks_notebook_operator.databricks_run_id = "1"
    databricks_notebook_operator.monitor_databricks_job()
    mock_runs_api.return_value.get_run.assert_called_with(
        databricks_notebook_operator.databricks_run_id, version="2.1"
    )
    assert 'Check the job run in Databricks: https://databricks-instance-xyz.cloud.databricks.com/#job/1234/run/1' in caplog.messages


@mock.patch("astro_databricks.operators.notebook.DatabricksHook")
@mock.patch("astro_databricks.operators.notebook.RunsApi")
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_api_client"
)
@mock.patch(
    "astro_databricks.operators.notebook.DatabricksNotebookOperator._get_databricks_task_id"
)
def test_monitor_databricks_job_fail(
    mock_get_databricks_task_id,
    mock_get_api_client,
    mock_runs_api,
    mock_databricks_hook,
    databricks_notebook_operator,
):
    mock_get_databricks_task_id.return_value = "1"
    # Define the expected response
    response = {
        "run_page_url": "https://databricks-instance-xyz.cloud.databricks.com/#job/1234/run/1",
        "state": {
            "life_cycle_state": "TERMINATED",
            "result_state": "FAILED",
            "state_message": "job failed",
        },
        "tasks": [
            {
                "run_id": "1",
                "task_key": "1",
            }
        ],
    }
    mock_runs_api.return_value.get_run.return_value = response

    databricks_notebook_operator.databricks_run_id = "1"
    with pytest.raises(AirflowException):
        databricks_notebook_operator.monitor_databricks_job()



"""DatabricksWorkflowTaskGroup for submitting jobs to Databricks."""
from __future__ import annotations

import logging
from operator import itemgetter
from typing import Any

from airflow.configuration import conf
from airflow.models import BaseOperator, BaseOperatorLink
from airflow.models.dag import DAG, clear_task_instances
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance, TaskInstanceKey
from airflow.models.xcom import XCom
from airflow.plugins_manager import AirflowPlugin
from airflow.providers.databricks.hooks.databricks import DatabricksHook
from airflow.security import permissions
from airflow.version import version as airflow_version
from packaging import version

try:
    # The following utility was included in Airflow version 2.3.3; we handle the needed import in the exception block.
    from airflow.utils.airflow_flask_app import get_airflow_app
except ModuleNotFoundError:
    # For older versions of airflow < 2.3.3 that don't have the utility.
    from flask import current_app

from airflow.exceptions import TaskInstanceNotFound
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.session import NEW_SESSION, provide_session
from airflow.utils.task_group import TaskGroup
from airflow.www.auth import has_access
from airflow.www.views import AirflowBaseView
from databricks_cli.sdk import JobsService
from databricks_cli.sdk.api_client import ApiClient
from flask import flash, redirect, request
from flask_appbuilder.api import expose
from sqlalchemy.orm.session import Session


def _get_flask_app():
    """Get the Airflow flask app instance"""
    try:
        flask_app = get_airflow_app()
    except NameError:
        flask_app = current_app
    return flask_app


def _get_databricks_task_id(task: BaseOperator) -> str:
    """Get the databricks task ID using dag_id and task_id. removes illegal characters.

    :param task: The task to get the databricks task ID for.
    :return: The databricks task ID.
    """
    return task.dag_id + "__" + task.task_id.replace(".", "__")


def get_databricks_task_ids(
    group_id: str, task_map: dict[str, BaseOperator], log: logging.Logger
) -> list[str]:
    """
    Returns a list of all Databricks task IDs for a dictionary of Airflow tasks.

    :param group_id: The task group ID.
    :param task_map: A dictionary mapping task IDs to BaseOperator instances.
    :return: A list of Databricks task IDs for the given task group.
    """
    task_ids = []
    log.debug("Getting databricks task ids for group %s", group_id)
    for task_id, task in task_map.items():
        if task_id == f"{group_id}.launch":
            continue
        databricks_task_id = _get_databricks_task_id(task)
        log.debug("databricks task id for task %s is %s", task_id, databricks_task_id)
        task_ids.append(databricks_task_id)
    return task_ids


@provide_session
def _get_dagrun(dag: DAG, run_id, session=None) -> DagRun:
    """
    Retrieves the DagRun object associated with the specified DAG and run_id.

    :param dag: The DAG object associated with the DagRun to retrieve.
    :param run_id: The run_id associated with the DagRun to retrieve.
    :param session: The SQLAlchemy session to use for the query. If None, uses the default session.
    :return: The DagRun object associated with the specified DAG and run_id.
    """
    return (
        session.query(DagRun)
        .filter(DagRun.dag_id == dag.dag_id, DagRun.run_id == run_id)
        .first()
    )


@provide_session
def _clear_task_instances(
    dag_id: str, run_id: str, task_ids: list[str], log: logging.Logger, session=None
):
    dag = _get_flask_app().dag_bag.get_dag(dag_id)
    log.debug("task_ids to clear", str(task_ids))
    dr: DagRun = _get_dagrun(dag, run_id)
    tis_to_clear = [
        ti for ti in dr.get_task_instances() if _get_databricks_task_id(ti) in task_ids
    ]
    clear_task_instances(tis_to_clear, session)


def _repair_task(
    databricks_conn_id: str,
    databricks_run_id: str,
    tasks_to_repair: list[str],
    log: logging.Logger,
) -> dict:
    """
    This function allows the Airflow retry function to create a repair job for Databricks.
    It uses the Databricks API to get the latest repair ID before sending the repair query.

    Note that we use the `JobsService` class instead of the `RunsApi` class. This is because the
    `RunsApi` class does not allow sending the `include_history` parameter which is necessary for
    repair jobs.

    Also for the moment we don't allow custom retry_callbacks. We might implement this in
    the future if users ask for it, but for the moment we want to keep things simple while the API
    stabilizes.

    :param databricks_conn_id: The Databricks connection ID.
    :param databricks_run_id: The Databricks run ID.
    :param tasks_to_repair: A list of Databricks task IDs to repair.
    :return: None
    """

    def _get_api_client():
        hook = DatabricksHook(databricks_conn_id)
        databricks_conn = hook.get_conn()
        return ApiClient(
            user=databricks_conn.login,
            token=databricks_conn.password,
            host=databricks_conn.host,
        )

    api_client = _get_api_client()
    log.debug("Getting latest repair ID")
    jobs_service = JobsService(api_client)
    current_job = jobs_service.get_run(run_id=databricks_run_id, include_history=True)
    repair_history = current_job.get("repair_history")
    repair_history_id = None
    if (
        repair_history and len(repair_history) > 1
    ):  # We use >1 because the first entry is the original run.
        # We use the last item in the array to get the latest repair ID
        repair_history_id = repair_history[-1]["id"]
        log.debug("Latest repair ID is %s", repair_history_id)
    log.debug(
        "Sending repair query for tasks %s on run %s",
        tasks_to_repair,
        databricks_run_id,
    )
    return jobs_service.repair(
        run_id=databricks_run_id,
        version="2.1",
        latest_repair_id=repair_history_id,
        rerun_tasks=tasks_to_repair,
    )


def get_task_group_legacy(operator: BaseOperator) -> TaskGroup:
    """
    Returns the task group for a given operator. This is a workaround for Airflow 2.2.4.
    Unfortunately in Airflow 2.2.4 the task_group property is not set on the operator, so we
    have to get the taskgroup tree from the DAG and search for the operator. This allows us to
    return the operators group, so we can find the xcom result of the launch task.

    :param operator: The operator to get the task group for.
    :return: The task group for the given operator.
    """

    def find_my_group(group: TaskGroup, task_id: str):
        groups_to_recurse = set()
        for elem in group.children.values():
            if isinstance(elem, TaskGroup):
                groups_to_recurse.add(elem)
            else:
                if operator.task_id == task_id:
                    return group
        for group in groups_to_recurse:
            val = find_my_group(group, task_id)
            if val:
                return val

    return find_my_group(operator.dag.task_group, operator.task_id)


def _get_launch_task_key(
    current_task_key: TaskInstanceKey, task_id: str
) -> TaskInstanceKey:
    """
    Returns the task key for the launch task. This allows us to gather databricks Metadata
    even if the current task has failed (since tasks only create xcom values if they succeed).

    :param current_task_key: The task key for the current task.
    :param task_id: The task ID for the current task.
    :return: The task key for the launch task.
    """
    if task_id:
        return TaskInstanceKey(
            dag_id=current_task_key.dag_id,
            task_id=task_id,
            run_id=current_task_key.run_id,
            try_number=current_task_key.try_number,
        )
    else:
        return current_task_key


@provide_session
def get_task_instance(operator, dttm, session: Session = NEW_SESSION):
    dag_id = operator.dag.dag_id
    dag_run = DagRun.find(dag_id, execution_date=dttm)[0]
    ti = (
        session.query(TaskInstance)
        .filter(
            TaskInstance.dag_id == dag_id,
            TaskInstance.run_id == dag_run.run_id,
            TaskInstance.task_id == operator.task_id,
        )
        .one_or_none()
    )
    if not ti:
        raise TaskInstanceNotFound("Task instance not found")
    return ti


def get_task_group(operator):
    if not hasattr(operator, "task_group"):
        task_group = get_task_group_legacy(operator)
    else:
        task_group = operator.task_group
    return task_group


def get_xcom_result(
    ti_key: TaskInstanceKey,
    key: str,
    ti: TaskInstance | None,
) -> Any:
    # Pull the xcom result for the given task instance and task instance key
    try:
        result = XCom.get_value(
            ti_key=ti_key,
            key=key,
        )
    except AttributeError:
        # For Airflow versions < 2.3.0 which do not contain the XCOM.get_value method implementation.
        if not ti:
            raise TaskInstanceNotFound(
                "Valid task instance is needed to fetch the XCOM result."
            )
        result = XCom.get_one(
            task_id=ti_key.task_id,
            dag_id=ti_key.dag_id,
            execution_date=ti.execution_date,
            key=key,
        )
    from astro_databricks.operators.workflow import DatabricksMetaData

    return DatabricksMetaData(**result)


class DatabricksJobRunLink(BaseOperatorLink, LoggingMixin):
    """Constructs a link to monitor a Databricks Job Run."""

    name = "See Databricks Job Run"

    def get_link(
        self,
        operator: BaseOperator,
        dttm=None,
        *,
        ti_key: TaskInstanceKey | None = None,
    ) -> str:
        ti = None
        if not ti_key:
            ti = get_task_instance(operator, dttm)
            ti_key = ti.key
        task_group = get_task_group(operator)

        dag = _get_flask_app().dag_bag.get_dag(ti_key.dag_id)
        dag.get_task(ti_key.task_id)
        self.log.info("Getting link for task %s", ti_key.task_id)
        if ".launch" not in ti_key.task_id:
            self.log.debug(
                "Finding the launch task for job run metadata %s", ti_key.task_id
            )
            launch_task_id = task_group.get_child_by_label("launch").task_id
            ti_key = _get_launch_task_key(ti_key, task_id=launch_task_id)
        # Should we catch the exception here if there is no return value?
        metadata = get_xcom_result(ti_key, "return_value", ti)

        hook = DatabricksHook(metadata.databricks_conn_id)
        return f"https://{hook.host}/#job/{metadata.databricks_job_id}/run/{metadata.databricks_run_id}"


class DatabricksJobRepairAllFailedLink(BaseOperatorLink, LoggingMixin):
    """Constructs a link to send a request to repair all failed databricks tasks."""

    name = "Repair All Failed Tasks"

    def get_link(
        self,
        operator,
        dttm=None,
        *,
        ti_key: TaskInstanceKey | None = None,
    ) -> str:
        ti = None
        if not ti_key:
            ti = get_task_instance(operator, dttm)
            ti_key = ti.key
        task_group = get_task_group(operator)
        self.log.debug(
            "Creating link to repair all tasks for databricks job run %s",
            task_group.group_id,
        )
        # Should we catch the exception here if there is no return value?
        metadata = get_xcom_result(ti_key, "return_value", ti)

        tasks_str = self.get_tasks_to_run(ti_key, operator, self.log)
        self.log.debug("tasks to rerun: %s", tasks_str)
        return (
            f"/repair_databricks_job?dag_id={ti_key.dag_id}&"
            f"databricks_conn_id={metadata.databricks_conn_id}&"
            f"databricks_run_id={metadata.databricks_run_id}&"
            f"run_id={ti_key.run_id}&"
            f"tasks_to_repair={tasks_str}"
        )

    def get_tasks_to_run(
        self, ti_key: TaskInstanceKey, operator: BaseOperator, log: logging.Logger
    ) -> str:
        task_group = get_task_group(operator)
        dag = _get_flask_app().dag_bag.get_dag(ti_key.dag_id)
        dr = _get_dagrun(dag, ti_key.run_id)
        log.debug("Getting failed and skipped tasks for dag run %s", dr.run_id)
        failed_and_skipped_tasks = self._get_failed_and_skipped_tasks(dr)
        log.debug("Failed and skipped tasks: %s", failed_and_skipped_tasks)
        tasks_to_run = {
            ti: t
            for ti, t in task_group.children.items()
            if ti in failed_and_skipped_tasks
        }
        log.debug(
            "Tasks to repair in databricks job %s : %s",
            task_group.group_id,
            tasks_to_run,
        )
        tasks_str = ",".join(
            get_databricks_task_ids(task_group.group_id, tasks_to_run, log)
        )
        return tasks_str

    def _get_failed_and_skipped_tasks(self, dr: DagRun) -> list[str]:
        """
        Returns a list of task IDs for tasks that have failed or have been skipped in the given DagRun.

        :param dr: The DagRun object for which to retrieve failed and skipped tasks.

        :return: A list of task IDs for tasks that have failed or have been skipped.
        """
        return [
            t.task_id
            for t in dr.get_task_instances(
                state=["failed", "skipped", "up_for_retry", "upstream_failed", None],
            )
        ]


class DatabricksJobRepairSingleFailedLink(BaseOperatorLink, LoggingMixin):
    """Constructs a link to send a repair request for a single databricks task."""

    name = "Repair a single failed task"

    def get_link(
        self,
        operator,
        dttm=None,
        *,
        ti_key: TaskInstanceKey | None = None,
    ) -> str:
        ti = None
        if not ti_key:
            ti = get_task_instance(operator, dttm)
            ti_key = ti.key

        task_group = get_task_group(operator)

        self.log.info(
            "Creating link to repair a single task for databricks job run %s task %s",
            task_group.group_id,
            ti_key.task_id,
        )
        dag = _get_flask_app().dag_bag.get_dag(ti_key.dag_id)
        task = dag.get_task(ti_key.task_id)
        # Should we catch the exception here if there is no return value?
        if ".launch" not in ti_key.task_id:
            launch_task_id = task_group.get_child_by_label("launch").task_id
            ti_key = _get_launch_task_key(ti_key, task_id=launch_task_id)
        metadata = get_xcom_result(ti_key, "return_value", ti)

        return (
            f"/repair_databricks_job?dag_id={ti_key.dag_id}&"
            f"databricks_conn_id={metadata.databricks_conn_id}&"
            f"databricks_run_id={metadata.databricks_run_id}&"
            f"tasks_to_repair={_get_databricks_task_id(task)}&"
            f"run_id={ti_key.run_id}"
        )


class RepairDatabricksTasks(AirflowBaseView, LoggingMixin):
    default_view = "repair"

    @expose("/repair_databricks_job", methods=["GET"])
    @has_access(
        [
            (permissions.ACTION_CAN_READ, permissions.RESOURCE_WEBSITE),
        ]
    )
    def repair(self):
        databricks_conn_id, databricks_run_id, dag_id, tasks_to_repair = itemgetter(
            "databricks_conn_id", "databricks_run_id", "dag_id", "tasks_to_repair"
        )(request.values)
        view = conf.get("webserver", "dag_default_view")
        return_url = self._get_return_url(dag_id, view)
        run_id = request.values.get("run_id").replace(
            " ", "+"
        )  # get run id separately since we need to modify it
        if tasks_to_repair == "":
            # If there are no tasks to repair, we return.
            flash("No tasks to repair. Not sending repair request.")
            return redirect(return_url)
        self.log.info("Repairing databricks job %s", databricks_run_id)
        res = _repair_task(
            databricks_conn_id=databricks_conn_id,
            databricks_run_id=databricks_run_id,
            tasks_to_repair=tasks_to_repair.split(","),
            log=self.log,
        )
        self.log.info(
            "Repairing databricks job query for run %s sent", databricks_run_id
        )
        self.log.info("Clearing tasks to rerun in airflow")
        _clear_task_instances(dag_id, run_id, tasks_to_repair.split(","), self.log)
        flash(f"Databricks repair job is starting!: {res}")
        return redirect(return_url)

    @staticmethod
    def _get_return_url(dag_id, view):
        if version.parse(airflow_version) < version.parse("2.3.0"):
            return_url = f"/{view}?dag_id={dag_id}"
        else:
            return_url = f"/dags/{dag_id}/{view}"
        return return_url


repair_databricks_view = RepairDatabricksTasks()

repair_databricks_package = {
    "name": "Repair Databricks View",
    "category": "Repair Databricks Plugin",
    "view": repair_databricks_view,
}


class AstroDatabricksPlugin(AirflowPlugin):
    name = "astro_databricks_plugin"
    operator_extra_links = [
        DatabricksJobRepairAllFailedLink(),
        DatabricksJobRepairSingleFailedLink(),
        DatabricksJobRunLink(),
    ]
    appbuilder_views = [repair_databricks_package]



"""DatabricksNotebookOperator for submitting notebook jobs to databricks."""
from __future__ import annotations

import time
from typing import Any

from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.providers.databricks.hooks.databricks import DatabricksHook
from airflow.utils.context import Context
from databricks_cli.runs.api import RunsApi
from databricks_cli.sdk.api_client import ApiClient

from astro_databricks.constants import JOBS_API_VERSION
from astro_databricks.operators.workflow import DatabricksMetaData
from astro_databricks.plugins.plugin import (
    DatabricksJobRepairSingleFailedLink,
    DatabricksJobRunLink,
)


class DatabricksNotebookOperator(BaseOperator):
    """
    Launches a notebook to databricks using an Airflow operator.

    The DatabricksNotebookOperator allows users to launch and monitor notebook
     deployments on Databricks as Aiflow tasks.
    It can be used as a part of a DatabricksWorkflowTaskGroup to take advantage of job clusters,
    which allows users to run their tasks on cheaper clusters that can be shared between tasks.

    Here is an example of running a notebook as a part of a workflow task group:

    .. code-block: python

        with dag:
            task_group = DatabricksWorkflowTaskGroup(
                group_id="test_workflow",
                databricks_conn_id="databricks_conn",
                job_clusters=job_cluster_spec,
                notebook_params=[],
            )
            with task_group:
                notebook_1 = DatabricksNotebookOperator(
                    task_id="notebook_1",
                    databricks_conn_id="databricks_conn",
                    notebook_path="/Users/daniel@astronomer.io/Test workflow",
                    source="WORKSPACE",
                    job_cluster_key="Shared_job_cluster",
                )
                notebook_2 = DatabricksNotebookOperator(
                    task_id="notebook_2",
                    databricks_conn_id="databricks_conn",
                    notebook_path="/Users/daniel@astronomer.io/Test workflow",
                    source="WORKSPACE",
                    job_cluster_key="Shared_job_cluster",
                    notebook_params={
                        "foo": "bar",
                    },
                )
                notebook_1 >> notebook_2


        :param notebook_path: the path to the notebook in Databricks
        :param source: Optional location type of the notebook. When set to WORKSPACE, the notebook will be retrieved
            from the local Databricks workspace. When set to GIT, the notebook will be retrieved from a Git repository
            defined in git_source. If the value is empty, the task will use GIT if git_source is defined
            and WORKSPACE otherwise. For more information please visit
            https://docs.databricks.com/dev-tools/api/latest/jobs.html#operation/JobsCreate
        :param databricks_conn_id: the connection id to use to connect to Databricks
        :param notebook_params: the parameters to pass to the notebook
    """

    operator_extra_links = (
        DatabricksJobRunLink(),
        DatabricksJobRepairSingleFailedLink(),
    )
    template_fields = ("databricks_metadata",)

    def __init__(
        self,
        notebook_path: str,
        source: str,
        databricks_conn_id: str,
        notebook_params: dict | None = None,
        notebook_packages: list[dict[str, Any]] = None,
        job_cluster_key: str | None = None,
        new_cluster: dict | None = None,
        existing_cluster_id: str | None = None,
        **kwargs,
    ):
        self.notebook_path = notebook_path
        self.source = source
        self.notebook_params = notebook_params or {}
        self.notebook_packages = notebook_packages or []
        self.databricks_conn_id = databricks_conn_id
        self.databricks_run_id = ""
        self.databricks_metadata: dict | None = None
        self.job_cluster_key = job_cluster_key or ""
        self.new_cluster = new_cluster or {}
        self.existing_cluster_id = existing_cluster_id or ""
        super().__init__(**kwargs)

        # For Airflow versions <2.3, the `task_group` attribute is unassociated, and hence we need to add that.
        if not hasattr(self, "task_group"):
            from airflow.utils.task_group import TaskGroupContext

            self.task_group = TaskGroupContext.get_current_task_group(self.dag)

    def _get_task_base_json(self) -> dict[str, Any]:
        """Get task base json to be used for task group tasks and single task submissions."""
        return {
            # Timeout seconds value of 0 for the Databricks Jobs API means the job runs forever.
            # That is also the default behavior of Databricks jobs to run a job forever without a default timeout value.
            "timeout_seconds": int(self.execution_timeout.total_seconds())
            if self.execution_timeout
            else 0,
            "email_notifications": {},
            "notebook_task": {
                "notebook_path": self.notebook_path,
                "source": self.source,
                "base_parameters": self.notebook_params,
            },
            "libraries": self.notebook_packages,
        }

    def convert_to_databricks_workflow_task(
        self, relevant_upstreams: list[BaseOperator]
    ):
        """
        Convert the operator to a Databricks workflow task that can be a task in a workflow
        """
        if hasattr(self.task_group, "notebook_packages"):
            self.notebook_packages.extend(self.task_group.notebook_packages)
        base_task_json = self._get_task_base_json()
        result = {
            "task_key": self._get_databricks_task_id(self.task_id),
            "depends_on": [
                {"task_key": self._get_databricks_task_id(t)}
                for t in self.upstream_task_ids
                if t in relevant_upstreams
            ],
            "job_cluster_key": self.job_cluster_key,
            **base_task_json,
        }
        return result

    def _get_databricks_task_id(self, task_id: str):
        """Get the databricks task ID using dag_id and task_id. removes illegal characters."""
        return self.dag_id + "__" + task_id.replace(".", "__")

    def monitor_databricks_job(self):
        """Monitor the Databricks job until it completes. Raises Airflow exception if the job fails."""
        api_client = self._get_api_client()
        runs_api = RunsApi(api_client)
        current_task = self._get_current_databricks_task(runs_api)
        url = runs_api.get_run(self.databricks_run_id, version=JOBS_API_VERSION)['run_page_url']
        self.log.info(f"Check the job run in Databricks: {url}")
        self._wait_for_pending_task(current_task, runs_api)
        self._wait_for_running_task(current_task, runs_api)
        self._wait_for_terminating_task(current_task, runs_api)
        final_state = runs_api.get_run(
            current_task["run_id"], version=JOBS_API_VERSION
        )["state"]
        self._handle_final_state(final_state)

    def _get_current_databricks_task(self, runs_api):
        return {
            x["task_key"]: x
            for x in runs_api.get_run(self.databricks_run_id, version=JOBS_API_VERSION)[
                "tasks"
            ]
        }[self._get_databricks_task_id(self.task_id)]

    def _handle_final_state(self, final_state):
        if final_state.get("life_cycle_state", None) != "TERMINATED":
            raise AirflowException(
                f"Databricks job failed with state {final_state}. Message: {final_state['state_message']}"
            )
        if final_state["result_state"] != "SUCCESS":
            raise AirflowException(
                "Task failed. Final State %s. Reason: %s",
                final_state["result_state"],
                final_state["state_message"],
            )

    def _get_lifestyle_state(self, current_task, runs_api):
        return runs_api.get_run(current_task["run_id"], version=JOBS_API_VERSION)[
            "state"
        ]["life_cycle_state"]

    def _wait_on_state(self, current_task, runs_api, state):
        while self._get_lifestyle_state(current_task, runs_api) == state:
            print(f"task {self.task_id.replace('.', '__')} {state.lower()}...")
            time.sleep(5)

    def _wait_for_terminating_task(self, current_task, runs_api):
        self._wait_on_state(current_task, runs_api, "TERMINATING")

    def _wait_for_running_task(self, current_task, runs_api):
        self._wait_on_state(current_task, runs_api, "RUNNING")

    def _wait_for_pending_task(self, current_task, runs_api):
        self._wait_on_state(current_task, runs_api, "PENDING")

    def _get_api_client(self):
        hook = DatabricksHook(self.databricks_conn_id)
        databricks_conn = hook.get_conn()
        return ApiClient(
            user=databricks_conn.login,
            token=databricks_conn.password,
            host=databricks_conn.host,
        )

    def launch_notebook_job(self):
        """Launch the notebook as a one-time job to Databricks."""
        api_client = self._get_api_client()
        base_task_json = self._get_task_base_json()
        run_json = {
            "run_name": self._get_databricks_task_id(self.task_id),
            **base_task_json,
        }
        if self.new_cluster and self.existing_cluster_id:
            raise ValueError(
                "Both new_cluster and existing_cluster_id are set. Only one can be set."
            )
        if self.existing_cluster_id:
            run_json["existing_cluster_id"] = self.existing_cluster_id
        elif self.new_cluster:
            run_json["new_cluster"] = self.new_cluster
        else:
            raise ValueError("Must specify either existing_cluster_id or new_cluster")
        runs_api = RunsApi(api_client)
        run = runs_api.submit_run(run_json)
        self.databricks_run_id = run["run_id"]
        return run

    def execute(self, context: Context) -> Any:
        """
        Execute the DataBricksNotebookOperator.

        Executes the DataBricksNotebookOperator. If the task is inside of a
        DatabricksWorkflowTaskGroup, it assumes the notebook is already launched
        and proceeds to monitor the running notebook.

        :param context:
        :return:
        """
        if not (
            hasattr(self.task_group, "is_databricks")
            and getattr(self.task_group, "is_databricks")
        ):
            self.launch_notebook_job()
        else:
            # if we are in a workflow, we assume there is a metadata from the launch task
            databricks_metadata = DatabricksMetaData(**self.databricks_metadata)
            self.databricks_run_id = databricks_metadata.databricks_run_id
            self.databricks_conn_id = databricks_metadata.databricks_conn_id
        self.monitor_databricks_job()



"""Example DAG for using the DatabricksNotebookOperator."""
import os
from datetime import timedelta

from airflow.models.dag import DAG
from airflow.utils.timezone import datetime

from astro_databricks import DatabricksNotebookOperator

EXECUTION_TIMEOUT = int(os.getenv("EXECUTION_TIMEOUT", 6))
default_args = {
    "execution_timeout": timedelta(hours=EXECUTION_TIMEOUT),
    "retries": int(os.getenv("DEFAULT_TASK_RETRIES", 2)),
    "retry_delay": timedelta(seconds=int(os.getenv("DEFAULT_RETRY_DELAY_SECONDS", 60))),
}

DATABRICKS_CONN_ID = os.getenv("ASTRO_DATABRICKS_CONN_ID", "databricks_conn")
NEW_CLUSTER_SPEC = {
    "cluster_name": "",
    "spark_version": "11.3.x-scala2.12",
    "aws_attributes": {
        "first_on_demand": 1,
        "availability": "SPOT_WITH_FALLBACK",
        "zone_id": "us-east-2b",
        "spot_bid_price_percent": 100,
        "ebs_volume_count": 0,
    },
    "node_type_id": "i3.xlarge",
    "spark_env_vars": {"PYSPARK_PYTHON": "/databricks/python3/bin/python3"},
    "enable_elastic_disk": False,
    "data_security_mode": "LEGACY_SINGLE_USER_STANDARD",
    "runtime_engine": "STANDARD",
    "num_workers": 8,
}

dag = DAG(
    dag_id="example_databricks_notebook",
    start_date=datetime(2022, 1, 1),
    schedule_interval=None,
    catchup=False,
    default_args=default_args,
    tags=["example", "async", "databricks"],
)
with dag:
    notebook_1 = DatabricksNotebookOperator(
        task_id="notebook_1",
        databricks_conn_id=DATABRICKS_CONN_ID,
        notebook_path="/Shared/Notebook_1",
        notebook_packages=[
            {
                "pypi": {
                    "package": "simplejson==3.18.0",
                    "repo": "https://pypi.org/simple",
                }
            },
            {"pypi": {"package": "Faker"}},
        ],
        source="WORKSPACE",
        new_cluster=NEW_CLUSTER_SPEC,
    )



"""Description of the package"""
from astro_databricks.operators.notebook import DatabricksNotebookOperator
from astro_databricks.operators.workflow import DatabricksWorkflowTaskGroup

__version__ = "0.1.1"
__all__ = [
    "DatabricksNotebookOperator",
    "DatabricksWorkflowTaskGroup"
]




from __future__ import annotations

from pathlib import Path

import pytest
from airflow.models.dagbag import DagBag
from airflow.utils.db import create_default_connections
from airflow.utils.session import provide_session

from . import utils as test_utils


@provide_session
def get_session(session=None):
    create_default_connections(session)
    return session


@pytest.fixture()
def session():
    return get_session()


def get_dag_bag() -> DagBag:
    """Create a DagBag by adding the files that are not supported to .airflowignore"""
    example_dags_dir = Path(__file__).parent.parent / "example_dags"

    dag_bag = DagBag(example_dags_dir, include_examples=False)
    assert dag_bag.dags
    assert not dag_bag.import_errors
    return dag_bag


dag_bag = get_dag_bag()


@pytest.mark.parametrize("dag_id", dag_bag.dag_ids)
def test_example_dag(session, dag_id: str):
    dag = dag_bag.get_dag(dag_id)
    test_utils.run_dag(dag, conn_file_path="test-connections.yaml")



"""DatabricksWorkflowTaskGroup for submitting jobs to Databricks."""
from __future__ import annotations

from logging import Logger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass
import json
import time
from typing import Any

from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.providers.databricks.hooks.databricks import DatabricksHook
from airflow.utils.context import Context
from airflow.utils.task_group import TaskGroup

try:
    from attrs import define
except ModuleNotFoundError:
    from attr import define

from databricks_cli.jobs.api import JobsApi
from databricks_cli.runs.api import RunsApi
from databricks_cli.sdk.api_client import ApiClient
from mergedeep import merge

from astro_databricks.plugins.plugin import (
    DatabricksJobRepairAllFailedLink,
    DatabricksJobRunLink,
)

@define
class DatabricksMetaData:
    databricks_conn_id: str
    databricks_run_id: str
    databricks_job_id: str


def _get_job_by_name(job_name: str, jobs_api: JobsApi) -> dict | None:
    jobs = jobs_api.list_jobs().get("jobs", [])
    for job in jobs:
        if job.get("settings", {}).get("name") == job_name:
            return job
    return None


class _CreateDatabricksWorkflowOperator(BaseOperator):
    """Creates a databricks workflow from a DatabricksWorkflowTaskGroup.

    :param task_id: The task id of the operator
    :param databricks_conn_id: The databricks connection id
    :param job_clusters: A list of job clusters to use in the workflow
    :param existing_clusters: A list of existing clusters to use in the workflow
    :param max_concurrent_runs: The maximum number of concurrent runs
    :param tasks_to_convert: A list of tasks to convert to a workflow. This list can also
    be populated after initialization by calling add_task.
    :param extra_job_params: A dictionary containing properties which will override the
    default Databricks Workflow Job definitions.
    """

    operator_extra_links = (DatabricksJobRunLink(), DatabricksJobRepairAllFailedLink())
    databricks_conn_id: str
    databricks_run_id: str
    databricks_job_id: str

    def __init__(
        self,
        task_id,
        databricks_conn_id,
        job_clusters: list[dict[str, object]] = None,
        existing_clusters: list[str] = None,
        max_concurrent_runs: int = 1,
        tasks_to_convert: list[BaseOperator] = None,
        extra_job_params: dict[str, Any] = None,
        **kwargs,
    ):
        self.existing_clusters = existing_clusters or []
        self.job_clusters = job_clusters or []
        self.job_cluster_dict = {j["job_cluster_key"]: j for j in self.job_clusters}
        self.tasks_to_convert = tasks_to_convert or []
        self.relevant_upstreams = [task_id]
        self.databricks_conn_id = databricks_conn_id
        self.databricks_run_id = None
        self.max_concurrent_runs = max_concurrent_runs
        self.extra_job_params = extra_job_params or {}
        super().__init__(task_id=task_id, **kwargs)

        # For Airflow versions <2.3, the `task_group` attribute is unassociated, and hence we need to add that.
        if not hasattr(self, "task_group"):
            from airflow.utils.task_group import TaskGroupContext

            self.task_group = TaskGroupContext.get_current_task_group(self.dag)

    def add_task(self, task: BaseOperator):
        """
        Add a task to the list of tasks to convert to a workflow.

        :param task:
        :return:
        """
        self.tasks_to_convert.append(task)

    def create_workflow_json(self) -> dict[str, object]:
        """Create a workflow json that can be submitted to databricks.

        :return: A workflow json
        """
        task_json = [
            task.convert_to_databricks_workflow_task(
                relevant_upstreams=self.relevant_upstreams
            )
            for task in self.tasks_to_convert
        ]
        default_json = {
            "name": self.databricks_job_name,
            "email_notifications": {"no_alert_for_skipped_runs": False},
            "timeout_seconds": 0,
            "tasks": task_json,
            "format": "MULTI_TASK",
            "job_clusters": self.job_clusters,
            "max_concurrent_runs": self.max_concurrent_runs,
        }
        merged_json = merge(default_json, self.extra_job_params)
        return merged_json

    @property
    def databricks_job_name(self):
        return self.dag_id + "." + self.task_group.group_id

    def execute(self, context: Context) -> Any:
        hook = DatabricksHook(self.databricks_conn_id)
        databricks_conn = hook.get_conn()
        api_client = ApiClient(
            token=databricks_conn.password, host=databricks_conn.host
        )
        jobs_api = JobsApi(api_client)
        job = _get_job_by_name(self.databricks_job_name, jobs_api)

        job_id = job["job_id"] if job else None
        current_job_spec = self.create_workflow_json()
        if not isinstance(self.task_group, DatabricksWorkflowTaskGroup):
            raise AirflowException("Task group must be a DatabricksWorkflowTaskGroup")
        if job_id:
            self.log.info(
                "Updating existing job with spec %s",
                json.dumps(current_job_spec, indent=4),
            )

            jobs_api.reset_job(
                json={"job_id": job_id, "new_settings": current_job_spec}
            )
        else:
            self.log.info(
                "Creating new job with spec %s", json.dumps(current_job_spec, indent=4)
            )
            job_id = jobs_api.create_job(json=current_job_spec)["job_id"]

        run_id = jobs_api.run_now(
            job_id=job_id,
            jar_params=self.task_group.jar_params,
            notebook_params=self.task_group.notebook_params,
            python_params=self.task_group.python_params,
            spark_submit_params=self.task_group.spark_submit_params,
        )["run_id"]
        runs_api = RunsApi(api_client)
        url = runs_api.get_run(run_id).get("run_page_url")
        self.log.info(f"Check the job run in Databricks: {url}")
        while runs_api.get_run(run_id)["state"]["life_cycle_state"] == "PENDING":
            print("job pending")
            time.sleep(5)
        self.databricks_run_id = run_id

        return {
            "databricks_conn_id": self.databricks_conn_id,
            "databricks_job_id": job_id,
            "databricks_run_id": run_id,
        }


class DatabricksWorkflowTaskGroup(TaskGroup):
    """
    A task group that takes a list of tasks and creates a databricks workflow.

    The DatabricksWorkflowTaskGroup takes a list of tasks and creates a databricks workflow
    based on the metadata produced by those tasks. For a task to be eligible for this
    TaskGroup, it must contain the ``convert_to_databricks_workflow_task`` method. If any tasks
    do not contain this method then the Taskgroup will raise an error at parse time.

    Here is an example of what a DAG looks like with a DatabricksWorkflowTaskGroup:

    .. code-block:: python

        job_clusters = [
            {
                "job_cluster_key": "Shared_job_cluster",
                "new_cluster": {
                    "cluster_name": "",
                    "spark_version": "11.3.x-scala2.12",
                    "aws_attributes": {
                        "first_on_demand": 1,
                        "availability": "SPOT_WITH_FALLBACK",
                        "zone_id": "us-east-2b",
                        "spot_bid_price_percent": 100,
                        "ebs_volume_count": 0,
                    },
                    "node_type_id": "i3.xlarge",
                    "spark_env_vars": {"PYSPARK_PYTHON": "/databricks/python3/bin/python3"},
                    "enable_elastic_disk": False,
                    "data_security_mode": "LEGACY_SINGLE_USER_STANDARD",
                    "runtime_engine": "STANDARD",
                    "num_workers": 8,
                },
            }
        ]

    with dag:
        task_group = DatabricksWorkflowTaskGroup(
            group_id="test_workflow",
            databricks_conn_id="databricks_conn",
            job_clusters=job_cluster_spec,
            notebook_params=[],
            notebook_packages=[
                {
                    "pypi": {
                        "package": "simplejson"
                    }
                },
            ]
        )
        with task_group:
            notebook_1 = DatabricksNotebookOperator(
                task_id="notebook_1",
                databricks_conn_id="databricks_conn",
                notebook_path="/Users/<user>/Test workflow",
                source="WORKSPACE",
                job_cluster_key="Shared_job_cluster",
                notebook_packages=[
                    {
                        "pypi": {
                            "package": "Faker"
                        }
                    }
                ]
            )
            notebook_2 = DatabricksNotebookOperator(
                task_id="notebook_2",
                databricks_conn_id="databricks_conn",
                notebook_path="/Users/<user>/Test workflow",
                source="WORKSPACE",
                job_cluster_key="Shared_job_cluster",
                notebook_params={
                    "foo": "bar",
                },
            )
            notebook_1 >> notebook_2

    With this example, Airflow will produce a job named <dag_name>.test_workflow that will
    run notebook_1 and then notebook_2. The job will be created in the databricks workspace
    if it does not already exist. If the job already exists, it will be updated to match
    the workflow defined in the DAG.

    To minimize update conflicts, we recommend that you keep parameters in the ``notebook_params`` of the
    ``DatabricksWorkflowTaskGroup`` and not in the ``DatabricksNotebookOperator`` whenever possible.
        This is because tasks in the
    ``DatabricksWorkflowTaskGroup`` are passed in at the job trigger time and do not modify the job definition

    :param group_id: The name of the task group
    :param databricks_conn_id: The name of the databricks connection to use
    :param job_clusters: A list of job clusters to use for this workflow.
    :param notebook_params: A list of notebook parameters to pass to the workflow.These parameters will be passed to
    all notebook tasks in the workflow.
    :param notebook_packages: A list of dictionary of Python packages to be installed. Packages defined at the
    workflow task group level are installed for each of the notebook tasks under it. And packages defined at the
    notebook task level are installed specific for the notebook task.
    :param jar_params: A list of jar parameters to pass to the workflow. These parameters will be passed to all jar
        tasks
    in the workflow.
    :param python_params: A list of python parameters to pass to the workflow. These parameters will be passed to
        all python tasks
    in the workflow.
    :param spark_submit_params: A list of spark submit parameters to pass to the workflow. These parameters
        will be passed to all spark submit tasks
    :param extra_job_params: A dictionary containing properties which will override the default Databricks Workflow
    Job definitions.
    :param max_concurrent_runs: The maximum number of concurrent runs for this workflow.

    """

    @property
    def log(self) -> Logger:
        """Returns logger."""
        pass

    is_databricks = True

    def __init__(
        self,
        databricks_conn_id,
        existing_clusters=None,
        job_clusters=None,
        jar_params: dict = None,
        notebook_params: list = None,
        notebook_packages: list[dict[str, Any]] = None,
        python_params: list = None,
        spark_submit_params: list = None,
        max_concurrent_runs: int = 1,
        extra_job_params: dict[str, Any] = None,
        **kwargs,
    ):
        """
        Create a new DatabricksWorkflowTaskGroup.

        :param group_id: The name of the task group
        :param databricks_conn_id: The name of the databricks connection to use
        :param job_clusters: A list of job clusters to use for this workflow.
        :param notebook_params: A list of notebook parameters to pass to the workflow.These parameters will be passed to
        all notebook tasks in the workflow.
        :param notebook_packages: A list of dictionary of Python packages to be installed. These packages will be passed
         to all notebook tasks in the workflow.
        :param jar_params: A list of jar parameters to pass to the workflow.
         These parameters will be passed to all jar tasks
        in the workflow.
        :param python_params: A list of python parameters to pass to the workflow.
         These parameters will be passed to all python tasks
        in the workflow.
        :param spark_submit_params: A list of spark submit parameters to pass to the workflow.
         These parameters will be passed to all spark submit tasks
        :param max_concurrent_runs: The maximum number of concurrent runs for this workflow.
        :param extra_job_params: A dictionary containing properties which will override the default Databricks
        Workflow Job definitions.
        """
        self.databricks_conn_id = databricks_conn_id
        self.existing_clusters = existing_clusters or []
        self.job_clusters = job_clusters or []
        self.notebook_params = notebook_params or []
        self.notebook_packages = notebook_packages or []
        self.python_params = python_params or []
        self.spark_submit_params = spark_submit_params or []
        self.jar_params = jar_params or []
        self.max_concurrent_runs = max_concurrent_runs
        self.extra_job_params = extra_job_params or {}
        super().__init__(**kwargs)

    def __exit__(self, _type, _value, _tb):
        """Exit the context manager and add tasks to a single _CreateDatabricksWorkflowOperator."""
        roots = self.roots

        # For Airflow versions <2.3, the `dag` attribute is unassociated, and hence we need to add that.
        if not hasattr(self, "dag"):
            from airflow.models.dag import DagContext

            self.dag = DagContext.get_current_dag()

        create_databricks_workflow_task: _CreateDatabricksWorkflowOperator = (
            _CreateDatabricksWorkflowOperator(
                dag=self.dag,
                task_id="launch",
                databricks_conn_id=self.databricks_conn_id,
                job_clusters=self.job_clusters,
                existing_clusters=self.existing_clusters,
                extra_job_params=self.extra_job_params,
            )
        )

        for task in roots:
            if not (
                hasattr(task, "convert_to_databricks_workflow_task")
                and callable(task.convert_to_databricks_workflow_task)
            ):
                raise AirflowException(
                    f"Task {task.task_id} does not support conversion to databricks workflow task."
                )
            create_databricks_workflow_task.set_upstream(
                task_or_task_list=list(task.upstream_list)
            )

        for task_id, task in self.children.items():
            if task_id != f"{self.group_id}.launch":
                create_databricks_workflow_task.relevant_upstreams.append(task_id)
                create_databricks_workflow_task.add_task(task)
                task.databricks_metadata = create_databricks_workflow_task.output

        create_databricks_workflow_task.set_downstream(roots)
        super().__exit__(_type, _value, _tb)



"""Nox automation definitions."""
import os
from pathlib import Path

import nox

nox.options.sessions = ["dev"]
nox.options.reuse_existing_virtualenvs = True


@nox.session(python="3.10")
def dev(session: nox.Session) -> None:
    """Create a dev environment with everything installed.

    This is useful for setting up IDE for autocompletion etc. Point the
    development environment to ``.nox/dev``.
    """
    session.install("nox")
    session.install("-e", ".[tests]")


def _expand_env_vars(file_path: Path):
    """Expand environment variables in the given file."""
    with file_path.open() as fp:
        yaml_with_env = os.path.expandvars(fp.read())
    with file_path.open("w") as fp:
        fp.write(yaml_with_env)


@nox.session(python=["3.8", "3.9", "3.10"])
@nox.parametrize("airflow", ["2.2.4", "2.3", "2.4", "2.5"])
def test(session: nox.Session, airflow) -> None:
    """Run both unit and integration tests."""
    env = {
        "AIRFLOW_HOME": f"~/airflow-{airflow}-python-{session.python}",
        "AIRFLOW__CORE__ALLOWED_DESERIALIZATION_CLASSES": "airflow\\.* astro\\.* astro_databricks\\.*",
    }

    session.install("-e", ".[tests]")
    session.install(f"apache-airflow=={airflow}")

    # Log all the installed dependencies
    session.log("Installed Dependencies:")
    session.run("pip3", "freeze")

    test_connections_file = Path("test-connections.yaml")
    if test_connections_file.exists():
        _expand_env_vars(test_connections_file)

    session.run("airflow", "db", "init", env=env)

    # Since pytest is not installed in the nox session directly, we need to set `external=true`.
    session.run(
        "pytest",
        "-vv",
        *session.posargs,
        env=env,
        external=True,
    )


@nox.session(python=["3.8"])
def type_check(session: nox.Session) -> None:
    """Run MyPy checks."""
    session.install("-e", ".[tests]")
    session.run("mypy", "--version")
    session.run("mypy", "src")


@nox.session(python="3.9")
def build_docs(session: nox.Session) -> None:
    """Build release artifacts."""
    session.install("-e", ".[docs]")
    session.chdir("./docs")
    session.run("make", "html")
